"use strict";

var Immutable = require("immutable");

var write = function write(state, tree) {
  return typeof(tree) === "string" ? state.update("code", function (code) {
    return code + convertValue(tree);
  }) : (function () {
    switch (tree.get(1)) {
      case "#include":
        return exports.include(state, tree);

      case "int":
        return exports.int(state, tree);

      case "float":
        return exports.float(state, tree);

      case "char":
        return exports.char(state, tree);

      case "string":
        return exports.string(state, tree);

      case "=":
        return exports.assign(state, tree);

      case "::":
        return exports.type(state, tree);

      case "\\":
        return exports.func(state, tree);

      case "return":
        return exports["return"](state, tree);

      case "+":
        return exports.add(state, tree);

      default:
        return exports.notHandled(state, tree);
    }
  })();
};

var convertValue = function convertValue(token) {
  return (function () {
    switch (true) {
      case token.match(/^-?[\d\.]+$/) != null:
        return token;

      case token.match(/^:/) != null:
        return JSON.stringify(token.substr(1));

      default:
        return token;
    }
  })();
};

exports.write = write;

exports.include = function (state, tree) {
  var first = tree.get(0);
  return state.update("code", function (code) {
    return code + "#include <" + first + ">";
  });
};

exports.int = function (state, tree) {
  var first = tree.get(0);
  return state.update("code", function (code) {
    return code + "int " + first;
  });
};

exports.char = function (state, tree) {
  var first = tree.get(0);
  return state.update("code", function (code) {
    return code + "char " + first;
  });
};

exports.string = function (state, tree) {
  var first = tree.get(0);
  return state.update("code", function (code) {
    return code + "char " + first + "[]";
  });
};

exports.float = function (state, tree) {
  var first = tree.get(0);
  return state.update("code", function (code) {
    return code + "float " + first + "[]";
  });
};

exports.assign = function (state, tree) {
  var leftValue = tree.get(0);
  var rightValue = tree.get(2);
  var state1 = exports.write(state, leftValue);
  var state2 = state1.update("code", function (code) {
    return code + " = ";
  });
  var state3 = exports.write(state2, rightValue);
  return state3.update("code", function (code) {
    return code + ";";
  });
};

exports["return"] = function (state, tree) {
  var first = tree.get(0);
  var state1 = state.update("code", function (code) {
    return code + "return ";
  });
  var state2 = exports.write(state1, first);
  return state2.update("code", function (code) {
    return code + ";";
  });
};

exports.notHandled = function (state, tree) {
  return state.update("code", function (code) {
    return code + "/* not" + JSON.stringify(tree) + "handled */";
  });
};

exports.func = function (state, tree) {
  var funcName = tree.get(0),
      args = tree.get(2),
      body = tree.slice(3),
      typeInfo = state.getIn(["types", funcName]);

  if (!(typeInfo != null)) {
    throw "type anotations not found for: " + funcName;
  }

  var argsCode = args.map(function (arg, index) {
    var itsType = typeInfo.getIn(["arguments", index]);
    return itsType + " " + arg;
  }).join(", ");
  var state1 = state.update("code", function (code) {
    return code + typeInfo.get("return") + " " + funcName + "(" + argsCode + ") {\n";
  }).update("indentation", function (indentation) {
    return indentation + "  ";
  });
  var result = body.reduce(function (acc, line) {
    var acc1 = acc.update("code", function (code) {
      return code + acc.get("indentation");
    });
    var acc2 = exports.write(acc1, line);
    return acc2.update("code", function (code) {
      return code + "\n";
    });
  }, state1);
  return result.update("code", function (code) {
    return code + "}";
  }).update("indentation", function (indentation) {
    return indentation.substr(2);
  });
};

exports.type = function (state, tree) {
  var funcName = tree.get(0);
  var argTypes = tree.get(2);
  var returnType = tree.get(3);
  return state.update("types", function (types) {
    return types.set(funcName, Immutable.fromJS({
      arguments: argTypes,
      "return": returnType
    }));
  });
};

exports.add = function (state, tree) {
  var leftValue = convertValue(tree.get(0));
  var rightValue = convertValue(tree.get(2));
  return state.update("code", function (code) {
    return code + leftValue + " + " + rightValue;
  });
};